/*

HXED, Version 2.00, a Hexadecimal Editor for Binary Files
Copyright 1991 Anthony F. Stuart, All rights reserved.

Date    Name	Change
910209	AFS	Initial release

*/


#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "sd.h"
#include "def.h"
#include "file.h"
#include "mcb.h"
#include "proto.h"


/* Initialize the file map list that contains a */
/* node for each block in the file. */

/* File map nodes are allocated individually since additional nodes */
/* are added as the file grows and freed as the file shrinks. */

init_file_map ()
{
	x_addr	 size;
	x_addr	 addr;
	int	 this_size;
	int	 rc;

	FMN	*this_fmn;
	FMN	*prev_fmn;

	struct stat	 sb;

	/* Get file information */

	rc = fstat (mcb->base_fd, &sb);
	if (rc == ERR)
		fatal ("Cannot stat base file");

	/* Special case of an empty file is not currently handled */

	size = sb.st_size;
	if (size == 0)
		fatal ("File is empty");

	addr = 0;
	prev_fmn = (FMN *) NULL;

	/* Allocate a file map node for each block in the file */

	while (size > 0) {
		this_fmn = (FMN *) allocate (sizeof (FMN));
		this_size = min (size, BP_FILL_SZ);
		size -= this_size;
		this_fmn->addr = addr;
		this_fmn->next_fmn = (FMN *) NULL;
		this_fmn->prev_fmn = prev_fmn;
		this_fmn->buf = (x_char *) NULL;
		this_fmn->len = this_size;
		this_fmn->flags = BASE;
		if (prev_fmn)
			prev_fmn->next_fmn = this_fmn;
		else
			mcb->head_fmn = this_fmn;
		prev_fmn = this_fmn;
		addr += BP_FILL_SZ;
	}

	/* Initialize master control block with file information */

	mcb->len = sb.st_size;
	mcb->original_len = sb.st_size;
	mcb->base_mtime = sb.st_mtime;
	mcb->tail_fmn = prev_fmn;
	mcb->this_fmn = mcb->head_fmn;
	mcb->offset = 0;
	mcb->temp_addr = 0;
	mcb->change = FALSE;

	/* Start with a valid block, mcb->offset id guarenteed to be loaded */

	load_file_map_node ();
}

/* Reinitialize the file map after writing changes */
/* back to the original base file */

re_init_file_map ()
{
	FMN	*this_fmn;
	FMN	*next_fmn;

	/* Free the old file map */

	this_fmn = mcb->head_fmn;

	while (this_fmn) {
		next_fmn = this_fmn->next_fmn;
		free (this_fmn);
		this_fmn = next_fmn;
	}

	/* Reopen the file */

	close (mcb->base_fd);
	mcb->base_fd = open (mcb->base_path, O_BINARY|O_RDONLY);
	if (mcb->base_fd == ERR)
		fatal ("Cannot open base file");

	/* And initialize the file map */

	init_file_map ();
}

/* Initialize the pool of buffers that are used to store */
/* file blocks in memory when they are accessed */

init_buffer_pool ()
{
	int	 i;
	int	 memory_req;
	BPN	*this_bpn;
	BPN	*prev_bpn;

	/* Allocate one big chunk to be divided up */

	memory_req = mcb->bp_size * sizeof (BPN);
	mcb->head_bpn = (BPN *) allocate (memory_req);

	this_bpn = mcb->head_bpn;
	prev_bpn = (BPN *) NULL;

	/* Initialize each buffer pool node in the buffer pool list */

	for (i = 0; i < mcb->bp_size; i++) {
		this_bpn->age = 0;
		this_bpn->next_bpn = (BPN *) NULL;
		this_bpn->this_fmn = (FMN *) NULL;
		if (prev_bpn)
			prev_bpn->next_bpn = this_bpn;
		prev_bpn = this_bpn;
		this_bpn++;
	}
}

/* Reinitialize the buffer pool after writing changes */
/* back to the original base file */

re_init_buffer_pool ()
{
	BPN	*this_bpn;

	/* Mark each buffer pool node unused */

	this_bpn = mcb->head_bpn;

	while (this_bpn) {
		this_bpn->this_fmn = (FMN *) NULL;
		this_bpn->age = 0;
		this_bpn = this_bpn->next_bpn;
	}
}

/* Set the byte specified by offset to a new value */

/* assumptions: mcb->this_fmn->buf is guaranteed to be */
/* loaded and there is always at least one fmn per file. */

set_byte_value (offset, value)
x_addr	 offset;
int	 value;
{
	x_char	*byte_addr;

	/* Get a byte address and set file map node dirty flag */

	byte_addr = get_byte_addr (offset);
	mcb->this_fmn->flags |= DIRTY;

	/* Set file change flag and store the new value */

	mcb->change = TRUE;
	*byte_addr = value;
}

/* The get_byte_addr function returns the address */
/* of the character at offset bytes into the file. */

/* This address is not guaranteed to remain valid */
/* across subsequent calls. */

x_char *
get_byte_addr (offset)
x_addr	 offset;
{
	FMN	*this_fmn;
	FMN	*next_fmn;
	FMN	*prev_fmn;

	this_fmn = mcb->this_fmn;

	/* Check for address in current block */

	if ((offset >= curoff) && (offset < curend))
		return (curbuf() + offset - curoff);

	next_fmn = this_fmn->next_fmn;
	prev_fmn = this_fmn->prev_fmn;

	/* Check for address in next block */

	if (next_fmn && (offset >= curend) && (offset < nxtend)) {
		mcb->this_fmn = next_fmn;
		mcb->offset += this_fmn->len;
	}

	/* Check for address in previous block */

	else if (prev_fmn && (offset >= prvoff) && (offset < curoff)) {
		mcb->this_fmn = prev_fmn;
		mcb->offset -= prev_fmn->len;
	}

	/* Check for attempt to extend file */

	else if (offset >= maxlen)
		extend_file_map (offset);

	/* Start at beginning and look for block */

	else
		find_file_map_node (offset);

	/* Load buffer if necessary and return byte address */

	if (!in_core())
		load_file_map_node ();

	return (curbuf() + offset - curoff);
}

/* Starting at the beginning of the file, add the */
/* offset in each file map node until we find the */
/* file map node that corresponds to offset passed */
/* as an argument. */

/* Assumptions: offset is not beyond end of file */

find_file_map_node (offset)
x_addr	 offset;
{
	x_addr	 this_offset;
	x_addr	 last_offset;
	FMN	*this_fmn;
	FMN	*last_fmn;

	/* Start at the beginning */

	this_offset = 0;
	last_offset = 0;
	this_fmn = mcb->head_fmn;
	last_fmn = this_fmn;

	/* Since offset is in file, all nodes are guaranteed to exist */

	while (this_offset <= offset) {
		last_offset = this_offset;
		this_offset += this_fmn->len;
		last_fmn = this_fmn;
		this_fmn = this_fmn->next_fmn;
	}

	/* Update master control block with new offset and map file node */

	mcb->offset = last_offset;
	mcb->this_fmn = last_fmn;
}

/* Extend the file map to include the specified offset */

/* Note: offsets are relative to zero, lengths are relative to */
/* one. Add one when converting from offsets to lengths. */

extend_file_map (offset)
x_addr	 offset;
{
	int	 i;
	int	 new_len;
	x_addr	 tail_offset;
	FMN	*tail_fmn;
	FMN	*this_fmn;

	/* Start at the tail of the file map */

	tail_fmn = mcb->tail_fmn;
	tail_offset = mcb->len - tail_fmn->len;

	/* Add to last block if there is room */

	if (offset < (tail_offset + BP_BUF_SZ)) {
		mcb->this_fmn = tail_fmn;
		mcb->offset = tail_offset;
		if (!in_core())
			load_file_map_node ();
		new_len = (offset - tail_offset) + 1;
		for (i = tail_fmn->len; i < new_len; i++) {
			tail_fmn->buf[i] = mcb->def_fill;
		}
		tail_fmn->len = new_len;
		tail_fmn->flags |= DIRTY;
		mcb->len = offset + 1;		/* offset is zero relative */
		return;				/* length is one relative */
	}


	/* Otherwise add one or more blocks to the file map. They will */
	/* be initialized when paged in. */

	while (offset >= mcb->len) {
		this_fmn = (FMN *) allocate (sizeof (FMN));
		this_fmn->addr = 0;
		this_fmn->next_fmn = (FMN *) NULL;
		this_fmn->prev_fmn = mcb->tail_fmn;
		this_fmn->buf = (x_char *) NULL;
		this_fmn->len = BP_FILL_SZ;
		this_fmn->flags = NEW;
		mcb->tail_fmn->next_fmn = this_fmn;
		mcb->tail_fmn = this_fmn;
		mcb->len += BP_FILL_SZ;
	}

	/* Update master control block with new offset and map file node */

	mcb->offset = mcb->len - BP_FILL_SZ;
	mcb->this_fmn = mcb->tail_fmn;

	/* The last block is currently BP_FILL_SZ bytes x_addr. Shorten */
	/* it to make it just large enough to contain new address. */

	mcb->tail_fmn->len = (offset - mcb->offset) + 1;
	mcb->len = offset + 1;

}

/* Load mcb->this_fmn from base file, temp file or default fill */

load_file_map_node ()
{
	int	 rc;
	int	 fd;
	FMN	*this_fmn;
	BPN	*this_bpn;

	/* Get a buffer pool node and initialize it for this file map node */

	this_bpn = find_buffer_pool_node ();
	this_bpn->age = 0;
	this_bpn->this_fmn = mcb->this_fmn;

	/* Save the buffer addr and initialize the file map node */

	this_fmn = mcb->this_fmn;
	this_fmn->buf = this_bpn->buf;
	this_fmn->flags |= CORE;
	this_fmn->flags &= ~DIRTY;

	/* Fill on demand if this block is new (inserted or beyond eof) */

	if (this_fmn->flags & NEW) {
		memset (this_fmn->buf, mcb->def_fill, this_fmn->len);
		return;
	}

	/* Read from temp file if paged-out, else read from base file */

	if (this_fmn->flags & TEMP)
		fd = mcb->temp_fd;
	else
		fd = mcb->base_fd;

	/* Seek to correct slot in file */

	rc = lseek (fd, this_fmn->addr, 0);
	if (rc == ERR)
		fatal ("Seek error");

	/* Read slot from file into buffer */

	rc = read (fd, this_fmn->buf, this_fmn->len);
	if (rc == ERR)
		fatal ("Read error");
}

/* Find a free buffer if available, else steal the oldest one in use */

BPN *
find_buffer_pool_node ()
{
	int	 rc;
	int	 age;
	BPN	*this_bpn;
	BPN	*old_bpn;
	FMN	*this_fmn;

	/* Search for a free buffer or an old one */

	this_bpn = mcb->head_bpn;
	old_bpn = (BPN *) NULL;
	age = 0;

	while (this_bpn) {

		/* Return a free buffer if one is found */

		if (this_bpn->this_fmn == (FMN *) NULL)
			return (this_bpn);

		/* Otherwise check for old buffer that is not locked */

		if (!(this_bpn->this_fmn->flags & LOCKED) &&
			(++(this_bpn->age) > age)) {
				old_bpn = this_bpn;
				age = this_bpn->age;
		}

		/* Move on to next buffer in buffer pool */

		this_bpn = this_bpn->next_bpn;

	}

	/* Did not find an old buffer so steal the oldest one */
	/* First get file map node associated with buffer */
	/* Then detach buffer from map node by resetting the in core flag */

	this_fmn = old_bpn->this_fmn;
	this_fmn->flags &= ~CORE;

	/* Return buffer it has not been updated since last read */

	if ((this_fmn->flags & DIRTY) == FALSE)
		return (old_bpn);

	/* Allocate space on temp file if block has never been written to it */

	if ((this_fmn->flags & TEMP) == FALSE) {
		this_fmn->addr = mcb->temp_addr;
		mcb->temp_addr += BP_BUF_SZ;
		this_fmn->flags &= ~NEW;
		this_fmn->flags &= ~BASE;
		this_fmn->flags |= TEMP;
	}

	/* Position to slot in temp file and write modified block */

	rc = lseek (mcb->temp_fd, this_fmn->addr, 0);
	if (rc == ERR)
		fatal ("Temp file seek error");

	rc = write (mcb->temp_fd, old_bpn->buf, BP_BUF_SZ);
	if (rc == ERR)
		fatal ("Temp file write error");

	return (old_bpn);
}

/* Allocate memory and signal fatal error if out of memory */

char *
allocate (len)
int	 len;
{
	char	*ptr;

	ptr = malloc (len);
	if (ptr == (char *) NULL)
		fatal ("Could not allocate memory");

	return (ptr);
}

/* Insert a byte with value and offset specified by */
/* argument list. All bytes at offset and greater are */
/* shifted down one. */

insert_byte_value (offset, value)
x_addr	 offset;
int	 value;
{
	int	 i;
	int	 move_amount;
	int	 insert_location;
	x_char	*buf;
	FMN	*old_fmn;
	FMN	*new_fmn;
	FMN	*this_fmn;

	/* An insert beyond the end of the file is equivalent */
	/* to a set and can be handled elsewhere. */

	if (offset >= mcb->len) {
		set_byte_value (offset, value);
		return;
	}

	/* Find block that contains offset and load into memory */

	get_byte_addr (offset);

	/* If no room left in this node then split it into two */
	/* by inserting a new node following this node and then */
	/* copying the second half of the buffer in this */
	/* node into the new node. Then the new node is made this */
	/* node and the insert occurs just as though there had */
	/* been enough room. */

	if (mcb->this_fmn->len == BP_BUF_SZ) {

		/* Allocate a new node */

		old_fmn = mcb->this_fmn;
		new_fmn = (FMN *) allocate (sizeof (FMN));
		new_fmn->addr = 0;
		new_fmn->next_fmn = old_fmn->next_fmn;
		new_fmn->prev_fmn = old_fmn;
		new_fmn->buf = (x_char *) NULL;
		new_fmn->len = 0;
		new_fmn->flags = NEW;

		/* Update linkage from prev and next nodes */

		old_fmn->next_fmn = new_fmn;
		if (new_fmn->next_fmn)
			new_fmn->next_fmn->prev_fmn = new_fmn;
		else
			mcb->tail_fmn = new_fmn;

		/* Figure out how much to move and move mcb to new node */

		move_amount = old_fmn->len / MOVE_FACTOR;
		mcb->offset += old_fmn->len - move_amount;
		mcb->this_fmn = new_fmn;

		/* Need to lock old_fmn in memory so it doesn't */
		/* get paged to load the new node */

		old_fmn->flags |= LOCKED;
		load_file_map_node ();
		old_fmn->flags &= ~LOCKED;

		/* Move second half of this buffer into new node */

		new_fmn->len = move_amount;
		old_fmn->len -= move_amount;
		memcpy (new_fmn->buf, old_fmn->buf+old_fmn->len, move_amount);

		/* The new node has been modified so mark it dirty. The */
		/* old node has not been modified, only its length has */
		/* changed. If the insert occurs in the old node it */
		/* will be marked dirty below. */

		new_fmn->flags |= DIRTY;
		new_fmn->flags &= ~NEW;

		/* If the offset to insert is in the previous block */
		/* then make it the current block. */

		if (offset < mcb->offset) {
			mcb->this_fmn = old_fmn;
			mcb->offset -= old_fmn->len;
		}

	}

	this_fmn = mcb->this_fmn;
	buf = this_fmn->buf;

	/* Move remaining bytes in buffer down one */

	insert_location = offset - mcb->offset;
	for (i = this_fmn->len; i > insert_location; i--) {
		buf[i] = buf[i-1];
	}

	/* Now insert new byte value and mark buffer updated */

	buf[insert_location] = value;
	this_fmn->flags |= DIRTY;

	/* Increment length of this buffer and file, mark file changed */

	this_fmn->len++;
	mcb->len++;
	mcb->change = TRUE;
}

/* Delete a byte with offset specified by argument list. All */
/* bytes at offset and greater are shifted up one. */

delete_byte_addr (offset)
x_addr	 offset;
{
	int	 i;
	int	 delete_location;
	int	 new_block_end;
	FMN	*this_fmn;
	x_char	*buf;

	/* Nothing to delete beyond end of file */

	if (offset >= mcb->len)
		return;

	/* Make sure the block is loaded into a buffer */

	get_byte_addr (offset);
	this_fmn = mcb->this_fmn;
	buf = this_fmn->buf;

	/* Shift characters in buffer down to replace deleted character */

	delete_location = offset - mcb->offset;
	new_block_end = this_fmn->len - 1;

	for (i = delete_location; i < new_block_end; i++) {
		buf[i] = buf[i+1];
	}

	/* Update flags and length field to indicate character was deleted */

	this_fmn->flags |= DIRTY;
	this_fmn->len--;
	mcb->len--;
	mcb->change = TRUE;

	/* Free the node if it is empty and not the last one in the file */

	if ((this_fmn->len > 0) || (mcb->len == 0))
		return;

	/* First update linkage from previous node or list head */

	if (this_fmn->prev_fmn)
		this_fmn->prev_fmn->next_fmn = this_fmn->next_fmn;
	else
		mcb->head_fmn = this_fmn->next_fmn;

	/* Next update linkage from next node or list tail */

	if (this_fmn->next_fmn)
		this_fmn->next_fmn->prev_fmn = this_fmn->prev_fmn;
	else
		mcb->tail_fmn = this_fmn->prev_fmn;

	/* Find the bpn associated with this block and free it */

	free_linked_buffer ();
	free (this_fmn);

	/* Go to a safe location that is guarenteed to exist */

	mcb->this_fmn = mcb->head_fmn;
	mcb->offset = 0;

	/* Current buffer is always guaranteed to be in core */

	if (!in_core())
		load_file_map_node ();
}

/* Search the buffer pool for a buffer address equal to */
/* the buffer address of this file map node and free it if */
/* found. This is less efficient for this operation than */
/* putting the bpn address in the fmn but this operation */
/* occurs infrequently and therefore not having the bpn in */
/* the fmn is more efficient overall in terms of both time */
/* and storage. */

free_linked_buffer ()
{
	BPN	*this_bpn;
	x_char	*buf_addr;

	buf_addr = mcb->this_fmn->buf;

	for (this_bpn = mcb->head_bpn; this_bpn; this_bpn=this_bpn->next_bpn) {
		if (this_bpn->buf == buf_addr) {
			this_bpn->this_fmn = (FMN *) NULL;
			return;
		}
	}
}

/* Write the current contents of the file map to an */
/* output file specified by the argument list. */

write_file_map (out_path)
char	*out_path;
{
	int	 rc;
	int	 out_fd;
	char	 buf[BP_BUF_SZ];
	FMN	*this_fmn;

	/* Open the output file, create or truncate as necessary */

	out_fd = open (out_path, O_BINARY|O_CREAT|O_TRUNC|O_WRONLY, 0666);
	if (out_fd == ERR)
		error ("Cannot open file for output");

	/* For every file node in the file map list */

	for (this_fmn = mcb->head_fmn; this_fmn; this_fmn=this_fmn->next_fmn) {

		/* If buffer is in core then write it to new file */

		if (this_fmn->flags & CORE) {
			rc = write (out_fd, this_fmn->buf, this_fmn->len);
			if (rc != this_fmn->len)
				fatal ("Output file write error");
		}

		/* Else if it is on base file then transfer to new file */

		else if (this_fmn->flags & BASE) {
			rc = lseek (mcb->base_fd, this_fmn->addr, 0);
			if (rc == ERR)
				fatal ("Base file seek error");
			rc = read (mcb->base_fd, buf, this_fmn->len);
			if (rc != this_fmn->len)
				fatal ("Base file read error");
			rc = write (out_fd, buf, this_fmn->len);
			if (rc != this_fmn->len)
				fatal ("Output file write error");
		}

		/* Else if it is in temp file then transfer to new file */

		else if (this_fmn->flags & TEMP) {
			rc = lseek (mcb->temp_fd, this_fmn->addr, 0);
			if (rc == ERR)
				fatal ("Temp file seek error");
			rc = read (mcb->temp_fd, buf, this_fmn->len);
			if (rc != this_fmn->len)
				fatal ("Temp file read error");
			rc = write (out_fd, buf, this_fmn->len);
			if (rc != this_fmn->len)
				fatal ("Output file write error");
		}

		/* Else if it is beyond end of file then default fill */

		else if (this_fmn->flags & NEW) {
			memset (buf, mcb->def_fill, this_fmn->len);
			rc = write (out_fd, buf, this_fmn->len);
			if (rc != this_fmn->len)
				fatal ("Output file write error");
		}
	}

	/* Close the new output file */

	close (out_fd);
}


