/*

HXED, Version 2.00, a Hexadecimal Editor for Binary Files
Copyright 1991 Anthony F. Stuart, All rights reserved.

Date    Name	Change
910209	AFS	Initial release

*/


/* 

   INTRODUCTION

   Files are accessed as a string of bytes using the functions
   provided by file.c. The basic operations are to get or set the
   value of a byte and to insert or delete a byte. There are also
   operations to read and write a file. The complexity of file
   manipulation is hidden from the editor.

   The master control block (mcb) contains the address of all the
   structures and fields that desribe the file being edited.

   MAPPING A FILE

   There is a list, called the file map list, made up of nodes
   called file map nodes.

   Each file map node maps one block in the file.

   The file map node contains a length field that specifies
   the size of the block.

   Initially each file map node refers to a block from the
   original file, referred to as the base file.

   The blocks from the base file are not read into a file
   map node buffer until needed.

   FINDING A BLOCK

   The block that addresses a particular byte in the file may be
   may be found by starting at the beginning of the file map
   list and summing block sizes until the block containing
   the address is found.

   There are also short cuts for finding the previous and
   next blocks. The address of these blocks is stored in the
   current block. 

   Since most accesses are relative to the current block, these
   short cuts save a great deal of time.

   READING A BLOCK

   A block must be read from the base file in order to read a
   byte within the block.

   The address of the block in the base file is stored in the
   file map node.

   Reading the block is simple. A buffer is aquired (more on that
   later), the file pointer is positioned to the address in the file
   map node, and the block is read into the buffer.

   UPDATING A BLOCK

   Once a block is in memory it can be updated. There are three
   basic ways that it can be updated. These are:

          overwrite byte
          insert byte
          delete byte

   Overwriting a byte is easy. The byte in the buffer is overstored
   with a new value.

   Inserting or deleting bytes is a little trickier because it changes
   the length of the block.

   Inserting a byte causes all bytes beyond the insert point to be
   shifted down (address incremented by one). In addition the block
   size and file size is incremented.

   When the file is originally mapped, some space is left in the
   block to make room for insertions, i.e. the initial block size
   is less than the buffer size.

   If the buffer becomes full a new file map node is inserted in
   the file map list.

   Deleting a byte causes all bytes beyond the delete point to be
   shifted up (address decremented by one). In addition the block size
   and file size is decremented.

   If a buffer becomes empty the file map node is removed from the
   file map list.

   ACCESSING BYTES BEYOND END OF FILE

   An access beyond the end of file causes the file map list to
   be extended to include more file map nodes.

   The bytes between the old end of file and the new end of file
   are filled with a default fill value specified by the user.

   The blocks are only filled when a byte contained in one is
   actually accessed.

   This means that if the file is extended by an access 20k beyond
   the end of file, only the file map nodes are created and inserted
   in the file map list, but no buffers are allocated or initialized
   for the intervening bytes.

   SAVING AN UPDATED BLOCK

   Once a buffer has been updated it is saved in memory until the
   buffer must be used for some other block.

   When this happens the block is paged-out (written out) to a 
   temporary file.

   Blocks are written to the temporary file in the order they are
   updated, not necessarily their order in the base file.

   The next block address to use in the temporary file is stored
   in the master control block.

   The file map node is updated with address of the block in
   the temporary file and a flag that indicates the block is now
   in the temporary file.

   If a byte in the block must be accessed again it is read from
   the temporary file instead of the base file.

   FREEING A NON-UPDATED BLOCK

   A buffer that has not been updated, i.e. it has only been read,
   can be freed without paging it out to the temporary file.

   If it is accessed again it is read from the base file just as
   it was originally.

   WRITING CHANGES BACK OUT

   The changes are written back out by starting at the head of the
   file map list and writing each block to the new file.

   The block may have to be read from the base file or the temporary
   file if it is not already in a buffer.

   Or it may have to initialized with the default fill if was
   beyond the original end of file.

   The output is always written to a temporary file to prevent
   the chance of overwriting the base file.

   The temporary file is then renamed with the name specified
   by the user.

   If the output file is the same as the base file the file map list
   is re-initialized to map the new file.

   ALLOCATING A BUFFER FOR A BLOCK

   As stated earlier, a block must be read into a buffer before
   a byte in the block can be accessed. Where do the buffers come
   from?

   A finite number of buffers are allocated at initialization time. These
   buffers are kept in a list called the buffer pool.

   There are usually fewer buffers than blocks, but this may not be
   true for very small files.

   When a buffer is needed the buffer pool list is searched for either
   a free buffer (one that has never been used) or one that has not been
   used in a long time.

   A free buffer is easy to use. The block is just read into it as
   described above.

   An old buffer is a little harder to use. If it contains a block that
   has been updated the block must be written to the temporary file first.

   ADDITIONAL NOTES

   The current file location is always specified by mcb->offset.

   This offset is always in the block pointed to by mcb->this_fmn.

   This block is always loaded or in the process of being loaded.
   
*/

/* File Map Node */

FMN {
	x_addr	 addr;		/* disk address in base or temp file	*/
	FMN	*next_fmn;	/* next file map node in list		*/
	FMN	*prev_fmn;	/* previous file map node in list	*/
	x_char	*buf;		/* addr of buffer when paged in		*/
	short	 len;		/* length of buffer when paged in	*/
	short	 flags;		/* current buffer location		*/
};

/* flags indicate the status of the block described by the file map node */

#define BASE	 0x01		/* always read from base file		*/
#define TEMP	 0x02		/* always read/write to temp file	*/
#define CORE	 0x04		/* block is in memory			*/
#define NEW	 0x08		/* beyond base end of file		*/
#define DIRTY	 0x10		/* buffer has been updated		*/
#define LOCKED	 0x20		/* buffer is locked, do not free	*/

/* base and temp are mutually exclusive */

/* example of flags

after init map node
	BASE

after initial page in
	BASE|CORE

after free node if not modified
	BASE

after page out if modified
	TEMP

after page in if previously modified
	TEMP|CORE

after update if not previously modified
	BASE|CORE|DIRTY

after update if previous modified
	TEMP|CORE|DIRTY

*/


/* Buffer Pool Node */

BPN {
	BPN		*next_bpn;	/* nxt buffer pool node in list	*/
	FMN		*this_fmn;	/* null if node is unused	*/
	x_char		 buf[BP_BUF_SZ];/* buffer to hold file block	*/
	unsigned short	 age;		/* node age used for stealing	*/
};

#define curoff	(mcb->offset)		/* current offset into file	*/
#define curlen	(this_fmn->len)		/* length of current buffer	*/
#define curend	(curoff+curlen)		/* end offset of current buffer	*/
#define maxlen	(mcb->len)		/* length of file		*/
#define nxtend	(curend+next_fmn->len)	/* end offset of next buffer	*/
#define prvoff	(curoff-prev_fmn->len)	/* beg offset of prev buffer	*/

#define curbuf()	(mcb->this_fmn->buf) /* address of current buffer	*/
#define in_core()	(mcb->this_fmn->flags & CORE) /* buf is valid	*/


